import Flutter
import UIKit
import kameleoonClient

public class SwiftKameleoonClientFlutterPlugin: NSObject, FlutterPlugin {
    private var clients = [String : KameleoonClient]()
    private var channel : FlutterMethodChannel

    private init(channel: FlutterMethodChannel) {
        self.channel = channel
    }

    public static func register(with registrar: FlutterPluginRegistrar) {
        let channel = FlutterMethodChannel(name: "kameleoon_client_flutter", binaryMessenger: registrar.messenger())
        let instance = SwiftKameleoonClientFlutterPlugin(channel: channel)
        registrar.addMethodCallDelegate(instance, channel: channel)
    }

    public func handle(_ call: FlutterMethodCall, result: @escaping FlutterResult) {
        print(call.method)
        let arguments = call.arguments as? NSDictionary
        let siteCode = arguments?["siteCode"] as? String
        let visitorCode = arguments?["visitorCode"] as? String

        if (call.method.elementsEqual("initializeKameleoonClient")) {
            let kameleoonClient = KameleoonClientFactory.create(sitecode : siteCode!)
            clients[siteCode!] = kameleoonClient
            result("")
        }
        else if (call.method.elementsEqual("triggerExperiment")) {
            let experimentID = arguments?["experimentID"] as? Int
            do {
                try result(clients[siteCode!]?.triggerExperiment(visitorCode : visitorCode!, experimentID : experimentID!))
            } catch {
                result(catchError(error))
            }
        }
        else if (call.method.elementsEqual("flush")) {
            do {
                try clients[siteCode!]?.flush(visitorCode : visitorCode!)
                result("")
            } catch {
                result(catchError(error))
            }
        }
        else if (call.method.elementsEqual("trackConversion")) {
          let goalID = arguments?["goalID"] as? Int
          let revenue = arguments?["revenue"] as? Float
            do {
                try clients[siteCode!]?.trackConversion(visitorCode : visitorCode!, goalId : goalID!, revenue : revenue)
            } catch {
                result(catchError(error))
            }
            result("")
        }
        else if (call.method.elementsEqual("activateFeature")) {
          let featureKey = arguments?["featureKey"] as? String
          let featureID = arguments?["featureID"] as? Int
          do {
            let feature : Bool
            if (featureID != nil) {
                try feature = clients[siteCode!]?.activateFeature(visitorCode : visitorCode!, featureID : featureID!) == true
            }
            else {
                try feature = clients[siteCode!]?.activateFeature(visitorCode : visitorCode!, featureKey : featureKey!) == true
            }
            result(feature)
          } catch {
            result(catchError(error))
          }
        }
        else if (call.method.elementsEqual("obtainVisitorCode")) {
          let defaultVisitorCode = arguments?["defaultVisitorCode"] as? String
          if (defaultVisitorCode != nil) {
            result(defaultVisitorCode)
          } else {
            result(UUID().uuidString)
          }
        }
        else if (call.method.elementsEqual("obtainVariationAssociatedData")) {
          let variationID = arguments?["variationID"] as? Int
          do {
            let data : Any?
            try data = clients[siteCode!]?.obtainVariationAssociatedData(variationID : variationID!)

            let jsonData = try JSONSerialization.data(withJSONObject: data!, options: JSONSerialization.WritingOptions.prettyPrinted)
            let dataAsString = String(data: jsonData, encoding: String.Encoding.utf8)
            result(dataAsString)
          } catch {
              result(catchError(error))
          }
        }
        else if (call.method.elementsEqual("obtainFeatureVariable")) {
          let featureKey = arguments?["featureKey"] as? String
          let featureID = arguments?["featureID"] as? Int
          let variableKey = arguments?["variableKey"] as? String
            do {
                if (featureID != nil)
                {
                  try result(clients[siteCode!]?.obtainFeatureVariable(featureID : featureID!, variableKey : variableKey!))
                }
                else
                {
                  try result(clients[siteCode!]?.obtainFeatureVariable(featureKey : featureKey!, variableKey : variableKey!))
                }
            }
            catch {
                result(catchError(error))
            }
        }
        else if (call.method.elementsEqual("addData")) {
            let dataType = arguments?["dataType"] as? String
            do {
                switch dataType {
                case "CustomData":
                    let id = arguments?["id"] as? Int
                    let value = arguments?["value"] as? String
                    try clients[siteCode!]?.addData(visitorCode : visitorCode!, data : CustomData(id : id!, value : value!))
                case "PageView":
                    let url = arguments?["url"] as? String
                    let title = arguments?["title"] as? String
                    let referrer = arguments?["referrer"] as? Int
                    try clients[siteCode!]?.addData(visitorCode : visitorCode!, data : PageView(url : url!, title : title!, referrer : referrer!))
                case "Interest":
                    let index = arguments?["index"] as? Int
                    try clients[siteCode!]?.addData(visitorCode : visitorCode!, data : Interest(index : index!))
                case "Conversion":
                    let goalId = arguments?["goalId"] as? Int
                    let revenue = arguments?["revenue"] as? Float
                    let negative = arguments?["negative"] as? Bool
                    try clients[siteCode!]?.addData(visitorCode : visitorCode!, data : Conversion(goalId : goalId!, revenue : revenue!, negative : negative!))
                default:
                  print("Data type not compatible")
                }
            } catch {
                result(catchError(error))
            }
          result(true);
        }
        else if (call.method.elementsEqual("runWhenReady")) {
            let timeout = arguments?["arguments"] as? Int ?? 2_000
            clients[siteCode!]?.runWhenReady(timeoutMilliseconds: timeout) { success in
                if (success) {
                    print("ready")
                    self.channel.invokeMethod("setReady", arguments: true)
                }
                else {
                    print("not ready")
                    self.channel.invokeMethod("setReady", arguments: false)
                }
            }
        }
        result("")
    }
    
    private func catchError(_ error: Error) -> FlutterError {
        
        enum VisitorCodeError: String {
            
            case empty = "visitorCodeEmpty", tooLong = "visitorCodeTooLong"
        }
        
        let errorCode: String
        switch error {
        case let error as KameleoonError where error == KameleoonError.experimentConfigurationNotFound:
            errorCode = "ExperimentConfigurationNotFound"
        case let error as KameleoonError where error == KameleoonError.featureConfigurationNotFound:
            errorCode = "FeatureConfigurationNotFound"
        case let error as KameleoonError where error == KameleoonError.sdkNotReady:
            errorCode = "SDKNotReady"
        case let error as KameleoonError where error == KameleoonError.notTargeted:
            errorCode = "NotTargeted"
        case let error as KameleoonError where error == KameleoonError.notActivated:
            errorCode = "NotActivated"
        case let error as KameleoonError where error == KameleoonError.variationConfigurationNotFound:
            errorCode = "VariationConfigurationNotFound"
        case let error as KameleoonError where error ==  KameleoonError.visitorCodeNotValid(VisitorCodeError.empty.rawValue) || error ==  KameleoonError.visitorCodeNotValid(VisitorCodeError.tooLong.rawValue):
            errorCode = "VisitorCodeNotValid"
        default:
            errorCode = "ERROR"
        }
        return FlutterError.init(
            code: errorCode,
            message: error.localizedDescription,
            details: nil)
    }
}
