import 'package:flutter/services.dart';
import '../kameleoon_client.dart';

import '../kameleoon_data.dart';
import '../kameleoon_exception.dart';
import 'kameleoon_client_callback.dart';

class KameleoonClientImpl extends KameleoonClient {
  static const MethodChannel _channel =
      const MethodChannel("kameleoon_client_flutter");

  bool ready = false;
  List<KameleoonClientCallback> callbacks = [];
  String? siteCode;

  KameleoonClientImpl(String siteCode) {
    this.siteCode = siteCode;
    _channel.setMethodCallHandler(nativeMethodCallHandler);
    _channel.invokeMethod("initializeKameleoonClient",
        <String, dynamic>{"siteCode": this.siteCode});
  }

  Future<void> trackConversion(String visitorCode, int goalID,
      [double? revenue]) async {
    try {
      await _channel.invokeMethod("trackConversion", <String, dynamic>{
        "visitorCode": visitorCode,
        "goalID": goalID,
        "revenue": revenue,
        "siteCode": this.siteCode
      });
    } on PlatformException catch (e) {
      switch (e.code) {
        case "VisitorCodeNotValid":
          throw VisitorCodeNotValid(e.message);
        default:
          throw Exception(e);
      }
    }
  }

  Future<int?> triggerExperiment(String visitorCode, int experimentID) async {
    try {
      return await _channel.invokeMethod("triggerExperiment", <String, dynamic>{
        "visitorCode": visitorCode,
        "experimentID": experimentID,
        "siteCode": this.siteCode
      });
    } on PlatformException catch (e) {
      switch (e.code) {
        case "SDKNotReady":
          throw SDKNotReady(e.message);
        case "ExperimentConfigurationNotFound":
          throw ExperimentConfigurationNotFound(e.message);
        case "NotTargeted":
          throw NotTargeted(e.message);
        case "NotActivated":
          throw NotActivated(e.message);
        case "VisitorCodeNotValid":
          throw VisitorCodeNotValid(e.message);
        default:
          throw Exception(e);
      }
    }
  }

  Future<bool?> activateFeature(String visitorCode, var featureKeyOrID) async {
    try {
      if (featureKeyOrID is int) {
        return await _channel.invokeMethod("activateFeature", <String, dynamic>{
          "visitorCode": visitorCode,
          "featureID": featureKeyOrID,
          "siteCode": this.siteCode
        });
      } else {
        return await _channel.invokeMethod("activateFeature", <String, dynamic>{
          "visitorCode": visitorCode,
          "featureKey": featureKeyOrID,
          "siteCode": this.siteCode
        });
      }
    } on PlatformException catch (e) {
      switch (e.code) {
        case "SDKNotReady":
          throw SDKNotReady(e.message);
        case "FeatureConfigurationNotFound":
          throw FeatureConfigurationNotFound(e.message);
        case "NotTargeted":
          throw NotTargeted(e.message);
        case "NotActivated":
          throw NotActivated(e.message);
        case "VisitorCodeNotValid":
          throw VisitorCodeNotValid(e.message);
        default:
          throw Exception(e);
      }
    }
  }

  Future<String?> obtainVisitorCode([String? defaultVisitorCode]) async {
    return await _channel.invokeMethod("obtainVisitorCode", <String, dynamic>{
      "defaultVisitorCode": defaultVisitorCode,
      "siteCode": this.siteCode
    });
  }

  Future<dynamic> obtainVariationAssociatedData(int variationID) async {
    try {
      return await _channel.invokeMethod(
          "obtainVariationAssociatedData", <String, dynamic>{
        "variationID": variationID,
        "siteCode": this.siteCode
      });
    } on PlatformException catch (e) {
      switch (e.code) {
        case "VariationConfigurationNotFound":
          throw VariationConfigurationNotFound(e.message);
        default:
          throw Exception(e);
      }
    }
  }

  Future<Object?> obtainFeatureVariable(
      var featureKeyOrID, String variableKey) async {
    var parameters = <String, dynamic>{
      "siteCode": siteCode,
      "variableKey": variableKey
    };
    if (featureKeyOrID is int) {
      parameters.putIfAbsent("featureID", () => featureKeyOrID);
    } else {
      parameters.putIfAbsent("featureKey", () => featureKeyOrID);
    }
    return await _channel.invokeMethod("obtainFeatureVariable", parameters);
  }

  bool? isReady() {
    return this.ready;
  }

  Future<void> addData(String visitorCode, Data data) async {
    var parameters = <String, dynamic>{};
    parameters.putIfAbsent("visitorCode", () => visitorCode);
    parameters.putIfAbsent("dataType", () => data.dataType);
    parameters.putIfAbsent("siteCode", () => this.siteCode);
    switch (data.dataType) {
      case "CustomData":
        CustomData d = data as CustomData;
        parameters.putIfAbsent("id", () => d.id);
        parameters.putIfAbsent("value", () => d.value);
        break;
      case "PageView":
        PageView d = data as PageView;
        parameters.putIfAbsent("url", () => d.url);
        parameters.putIfAbsent("title", () => d.title);
        parameters.putIfAbsent("referrer", () => d.referrer);
        break;
      case "Interest":
        Interest d = data as Interest;
        parameters.putIfAbsent("index", () => d.index);
        break;
      case "Conversion":
        Conversion d = data as Conversion;
        parameters.putIfAbsent("goalId", () => d.goalId);
        parameters.putIfAbsent("revenue", () => d.revenue);
        parameters.putIfAbsent("negative", () => d.negative);
        break;
      case "Browser":
        Browser d = data as Browser;
        parameters.putIfAbsent("browser", () => d.browser);
        break;
    }
    try {
      await _channel.invokeMethod("addData", parameters);
    } on PlatformException catch (e) {
      switch (e.code) {
        case "VisitorCodeNotValid":
          throw VisitorCodeNotValid(e.message);
        default:
          throw Exception(e);
      }
    }
  }

  void runWhenReady(Function readyCallback, Function failCallback,
      [int timeout = 2000]) async {
    if (this.ready) {
      await readyCallback.call();
    } else {
      final callback = new KameleoonClientCallback(readyCallback, failCallback);
      this.callbacks.add(callback);
      await _channel.invokeMethod("runWhenReady",
          <String, dynamic>{"timeout": timeout, "siteCode": this.siteCode});
      Future.delayed(Duration(milliseconds: timeout),
          () => callbackAfterStatusChange(false, callback));
    }
  }

  Future<void> flush([String? visitorCode]) async {
    var parameters = <String, dynamic>{
      "siteCode": siteCode,
    };
    try {
      if (visitorCode != null) {
        parameters.putIfAbsent("visitorCode", () => visitorCode);
      }
      await _channel.invokeMethod("flush", parameters);
    } on PlatformException catch (e) {
      switch (e.code) {
        case "VisitorCodeNotValid":
          throw VisitorCodeNotValid(e.message);
        default:
          throw Exception(e);
      }
    }
  }

  void callbackAfterStatusChange(bool isReady,
      [KameleoonClientCallback? callback]) {
    List<int> idsToDelete = [];
    for (int i = 0; i < this.callbacks.length; i++) {
      if (isReady) {
        this.callbacks[i].readyCallback.call();
        idsToDelete.add(i);
      } else {
        if (this.callbacks[i] == callback) {
          callback?.failCallback?.call();
          idsToDelete.add(i);
        }
      }
    }

    for (int i = idsToDelete.length - 1; i >= 0; i--) {
      this.callbacks.removeAt(i);
    }
  }

  Future<void> nativeMethodCallHandler(MethodCall call) async {
    switch (call.method) {
      case "setReady":
        this.ready = call.arguments ?? this.ready;
        callbackAfterStatusChange(this.ready);
        break;
      default:
        break;
    }
  }
}
