import 'kameleoon_data.dart';

abstract class KameleoonClient {
  // To track conversion, use the trackConversion() method. This method requires visitorCode and goalID to track conversion on this particular goal. In addition, this method also accepts revenue as a third optional argument to track revenue. The visitorCode should be identical to the one that was used when triggering the experiment.
  // The trackConversion() method doesn't return any value. This method is non-blocking as the server call is made asynchronously.
  Future<void> trackConversion(String visitorCode, int goalID, [double? revenue]);

  // Running an A/B experiment on your Flutter application means bucketing your users into several groups (one per variation). The SDK takes care of this bucketing (and the associated reporting) automatically.
  // Triggering an experiment by calling the triggerExperiment() method will register a random variation for a given visitorCode. If this visitorCode is already associated with a variation (most likely the user had already been exposed to the experiment previously), then it will return the previous variation associated with a given experiment.
  Future<int?> triggerExperiment(String visitorCode, int experimentID);

  // To activate a feature toggle, call the activateFeature() method of our SDK.
  // This method takes a visitorCode and featureKey (or featureID) as mandatory arguments to check if the specified feature will be active for a given user.
  // If such a user has never been associated with this feature flag, the SDK returns a boolean value randomly (true if the user should have this feature or false if not). If a user with a given visitorCode is already registered with this feature flag, it will detect the previous featureFlag value.
  // You have to make sure that proper error handling is set up in your code as shown in the example to the right to catch potential exceptions.
  Future<bool?> activateFeature(String visitorCode, var featureKeyOrID);

  Future<String?> obtainVisitorCode([String? defaultVisitorCode]);

  // To retrieve JSON data associated with a variation, call the obtainVariationAssociatedData() method of our SDK. The JSON data usually represents some metadata of the variation, and can be configured on our web application interface or via our Automation API.
  // This method takes the variationID as a parameter and will return the data as a JSON object. It will throw an exception (KameleoonException.VariationConfigurationNotFound) if the variation ID is wrong or corresponds to an experiment that is not yet online.
  Future<dynamic> obtainVariationAssociatedData(int variationID);

  // To retrieve a feature variable, call the obtainFeatureVariable() method of our SDK. A feature variable can be changed easily via our web application.
  // This method takes two input parameters: featureKey and variableKey. It will return the data as an Object instance. Usually it should be casted to the expected type (the one defined on the web interface). It will throw an exception (KameleoonException.FeatureConfigurationNotFound) if the requested feature has not been found in the internal configuration of the SDK.
  Future<Object?> obtainFeatureVariable(var featureKeyOrID, String variableKey);

  // For mobile SDKs, the initialization of the Kameleoon Client is not immediate, as it needs to perform a server call to retrieve the current configuration for all active experiments. It is recommended to check if the SDK is ready by calling this method before triggering an experiment. Alternatively, you could setup exception catching around the triggerExperiment() method to look for the KameleoonException.SDKNotReady exception, or you could use the runWhenReady() method with a callback as detailed in the next paragraph.
  bool? isReady();

  // To associate various data with the current user, we can use the addData() method. This method requires the visitorCode as a first parameter, and then accept several additional parameters. Those additional parameters represent the various Data Types allowed in Kameleoon.
  // Note that the addData() method doesn't return any value and doesn't interact with the Kameleoon back-end servers by itself. Instead, every data declared is saved for further sending via the flush() method described in the next paragraph. This allows a minimal number of server calls to be made, as data are usually regrouped in a single server call triggered by the execution of flush().
  Future<void> addData(String visitorCode, Data data);

  // For mobile SDKs, the initialization of the Kameleoon Client is not immediate, as it needs to perform a server call to retrieve the current configuration for all active experiments. The runWhenReady() method of the KameleoonClient class allows to pass a callback that will be executed as soon as the SDK is ready for use. It also allows the use of a timeout.
  // The first callback will be executed once the Kameleoon client is ready, and should contain code triggering an experiment and implementing variations. The second callback will be executed if the specified timeout happens before the client is initialized. Usually this case should implement the "reference" variation, as the user will be "out of the experiment" if a timeout takes place.
  void runWhenReady(Function readyCallback, Function failCallback, [int timeout = 2000]);

  // Data associated with the current user via addData() method is not sent immediately to the server. It is stored and accumulated until it is sent automatically by the triggerExperiment() or trackConversion() methods, or manually by the flush() method. This allows the developer to exactly control when the data is flushed to our servers. For instance, if you call the addData() method a dozen times, it would be a waste of ressources to send data to the server after each addData() invocation. Just call flush() once at the end.
  // The flush() method doesn't return any value. This method is non-blocking as the server call is made asynchronously.
  Future<void> flush([String? visitorCode]);
}
